@@ -25,6 +25,9 @@ from ema_pytorch import EMA
 
 from accelerate import Accelerator
 
+from pytorch_fid.inception import InceptionV3
+from pytorch_fid.fid_score import calculate_frechet_distance
+
 from denoising_diffusion_pytorch.version import __version__
 
 # constants
@@ -460,6 +463,7 @@ class GaussianDiffusion(nn.Module):
         assert not model.random_or_learned_sinusoidal_cond
 
         self.model = model
+
         self.channels = self.model.channels
         self.self_condition = self.model.self_condition
 
@@ -814,12 +818,15 @@ class Trainer(object):
         results_folder = './results',
         amp = False,
         fp16 = False,
-        use_lion = False,
         split_batches = True,
-        convert_image_to = None
+        convert_image_to = None,
+        calculate_fid = True,
+        inception_block_idx = 2048
     ):
         super().__init__()
 
+        # accelerator
+
         self.accelerator = Accelerator(
             split_batches = split_batches,
             mixed_precision = 'fp16' if fp16 else 'no'
@@ -827,8 +834,22 @@ class Trainer(object):
 
         self.accelerator.native_amp = amp
 
+        # model
+
         self.model = diffusion_model
 
+        # InceptionV3 for fid-score computation
+
+        self.inception_v3 = None
+
+        if calculate_fid:
+            assert inception_block_idx in InceptionV3.BLOCK_INDEX_BY_DIM
+            block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[inception_block_idx]
+            self.inception_v3 = InceptionV3([block_idx])
+            self.inception_v3.to(self.device)
+
+        # sampling and training hyperparameters
+
         assert has_int_squareroot(num_samples), 'number of samples must have an integer square root'
         self.num_samples = num_samples
         self.save_and_sample_every = save_and_sample_every
@@ -849,13 +870,13 @@ class Trainer(object):
 
         # optimizer
 
-        optim_klass = Lion if use_lion else Adam
-        self.opt = optim_klass(diffusion_model.parameters(), lr = train_lr, betas = adam_betas)
+        self.opt = Adam(diffusion_model.parameters(), lr = train_lr, betas = adam_betas)
 
         # for logging results in a folder periodically
 
         if self.accelerator.is_main_process:
             self.ema = EMA(diffusion_model, beta = ema_decay, update_every = ema_update_every)
+            self.ema.to(self.device)
 
         self.results_folder = Path(results_folder)
         self.results_folder.mkdir(exist_ok = True)
@@ -868,6 +889,10 @@ class Trainer(object):
 
         self.model, self.opt = self.accelerator.prepare(self.model, self.opt)
 
+    @property
+    def device(self):
+        return self.accelerator.device
+
     def save(self, milestone):
         if not self.accelerator.is_local_main_process:
             return
@@ -902,6 +927,25 @@ class Trainer(object):
         if exists(self.accelerator.scaler) and exists(data['scaler']):
             self.accelerator.scaler.load_state_dict(data['scaler'])
 
+    @torch.no_grad()
+    def calculate_activation_statistics(self, samples):
+        features = self.inception_v3(samples)[0]
+        features = rearrange(features, '... 1 1 -> ...')
+
+        mu = torch.mean(features, dim = 0).cpu()
+        sigma = torch.cov(features).cpu()
+        return mu, sigma
+
+    def fid_score(self, real_samples, fake_samples):
+        min_batch = min(real_samples.shape[0], fake_samples.shape[0])
+        real_samples, fake_samples = map(lambda t: t[:min_batch], (real_samples, fake_samples))
+
+        m1, s1 = self.calculate_activation_statistics(real_samples)
+        m2, s2 = self.calculate_activation_statistics(fake_samples)
+
+        fid_value = calculate_frechet_distance(m1, s1, m2, s2)
+        return fid_value
+
     def train(self):
         accelerator = self.accelerator
         device = accelerator.device
@@ -934,7 +978,6 @@ class Trainer(object):
 
                 self.step += 1
                 if accelerator.is_main_process:
-                    self.ema.to(device)
                     self.ema.update()
 
                     if self.step != 0 and self.step % self.save_and_sample_every == 0:
@@ -946,9 +989,16 @@ class Trainer(object):
                             all_images_list = list(map(lambda n: self.ema.ema_model.sample(batch_size=n), batches))
 
                         all_images = torch.cat(all_images_list, dim = 0)
+
                         utils.save_image(all_images, str(self.results_folder / f'sample-{milestone}.png'), nrow = int(math.sqrt(self.num_samples)))
                         self.save(milestone)
 
+                        # whether to calculate fid
+
+                        if exists(self.inception_v3):
+                            fid_score = self.fid_score(real_samples = data, fake_samples = all_images)
+                            accelerator.print(f'fid_score: {fid_score}')
+
                 pbar.update(1)
 
         accelerator.print('training complete')
